
//Consider a simple binary tree with 10 nodes with root node as A.

/*			A
		       / \
		      /   \
		     B     C
		    / \   / \
		   D   E F   G
		  / \  |
		 H   I J
*/
/* The following foo() method gives the high level idea of the implementation of Intention locks(IL) and DomLock.
   For the programmer's perspective, there is no difference as LockAll() method gives the higher level abstraction
   for synchronization. The implementation of LockAll differs for IL and DomLock as mention below. Consider that the
   locking library provides the interface for modes of locks like IX, X, IS, S. We provide same interface for FindDominator(), Lock()
   and UnLock() method used in DomLock version of LockAll() method.
*/
  
void foo()
{
	//before starting the critical section, the transaction request locks on nodes H and J in X-mode, i.e., exclusive mode.
	node* H,J;
	LockAll(X,H,J);
	//Critical section entry-point
	{
	
		//Perform some operation on H and J.
	}
	//Critical section exit-point
	UnlockAll(H,J);
	


}


//********************************************************************
//Implementation of LockAll method for Intention Locks
//********************************************************************

void LockAll(X,H,J)
{
    // Lock parent of H and J in intention exclusive mode.

	//All parents of H
	IX(A);
	IX(B);
	IX(D);
	//All parents of J
	IX(A);
	IX(B);
	IX(E);
   // Lock H and J in exclusive mode.

	X(H);
	X(J);

}

//********************************************************************
//Implementation of UnlockAll method for Intention Locks
//********************************************************************

void UnLockAll(X,H,J)
{


   // Unlock H and J in exclusive mode.

	UnLockX(H);
	UnLockX(J);

    // Unlock parent of H in intention exclusive mode in reverse order.
	UnlockIX(D);
	UnlockIX(B);
	UnlockIX(A);
   // Unlock parent of J in intention exclusive mode in reverse order.
	UnLockIX(E);
	UnlockIX(B);
	UnLockIX(A);
	

}


//********************************************************************
//Implementation of LockAll method for DomLock
//********************************************************************

void LockAll(X,H,J)
{
    //Find dominator for H and J.

	node* Dom = FindDominator(H,J);
	
   // Lock the dominator
	Lock(X,Dom);

}

//********************************************************************
//Implementation of UnLockAll method for DomLock
//********************************************************************
void UnLockAll(X,Dom)
{

	//UnLock the dominator node.
	UnLock(Dom);
}

